@@ -8,6 +8,8 @@ from uuid import UUID
 from grpc import RpcError
 
 from preseem import NetworkMetadataReference
+from preseem.source_model import Attachment, Service
+from preseem_protobuf.model.service_pb2 import SetAttachmentReq
 
 # STM-6039 Only allow the sync to run once at a time, to prevent duplicate
 # entities from being created.
@@ -40,6 +42,15 @@ _package_field_map = {
     'upload_rate': lambda x: x.upload_rate or None,
 }
 
+_service_field_map = {
+    'account_uuid': lambda x: x._account_uuid,
+    'package_uuid': lambda x: x._package_uuid,
+    'download_rate': lambda x: x._download_rate or None,
+    'upload_rate': lambda x: x._upload_rate or None,
+    'attachment': lambda x: x._attachment.pb if x._attachment else None,
+    'subscriber_identifier': lambda x: x._subscriber_identifier,
+}
+
 
 async def call_and_log_error(co):
     """Call a GRPC coroutine and log any errors from it."""
@@ -70,6 +81,7 @@ async def _sync_objects(model, company_uuid, source, bss_objs, model_objs, creat
       returns the value to sync from the bss object.
     """
     new_ids = set(bss_objs)
+    new_objs = []
     for mobj in model_objs:
         if mobj.source != source:
             continue  # ignore objects owned by other sources
@@ -85,12 +97,34 @@ async def _sync_objects(model, company_uuid, source, bss_objs, model_objs, creat
                 mval = getattr(mobj, f)
                 bval = m(bobj)
                 if mval != bval:
-                    changes.append(f'{f}: {mval} -> {bval}')
+                    if f == 'attachment':
+                        # find what attachment fields changed for logging purposes
+                        achgs = []
+                        if mval and bval:
+                            for sf in bval.DESCRIPTOR.fields:
+                                smval = getattr(mval, sf.name)
+                                sbval = getattr(bval, sf.name)
+                                if smval != sbval:
+                                    achgs.append(f"{sf.name}[{smval}->{sbval}]")
+                        changes.append(f'{f}: {",".join(achgs)}')
+                    else:
+                        changes.append(f'{f}: {mval} -> {bval}')
                     mods[f] = bval
             if mods:
-                logging.info('sync_preseem_model(%s): modify %s (%s)', source, mobj,
+                objdesc = mobj
+                if isinstance(bobj, Service):
+                    # more concise error message for Service
+                    objdesc = f"Service {UUID(bytes=mobj.uuid)}"
+                logging.info('sync_preseem_model(%s): modify %s (%s)', source, objdesc,
                              ', '.join(changes))
                 if not noact:
+                    if isinstance(bobj, Service) and 'attachment' in mods:
+                        # special case for setting attachment
+                        attachment = mods.pop('attachment')
+                        set_att_req = SetAttachmentReq()
+                        if attachment:
+                            set_att_req.attachment.CopyFrom(attachment)
+                        mods['attachment'] = set_att_req
                     await call_and_log_error(
                         updaterpc(company_uuid=company_uuid, uuid=mobj.uuid, **mods))
         else:
@@ -119,6 +153,8 @@ async def _sync_objects(model, company_uuid, source, bss_objs, model_objs, creat
                     uuid = UUID(bytes=mobj.uuid)
                     logging.info('sync_preseem_model(%s): created %s "%s"', source,
                                  type(mobj).__name__, uuid)
+                    new_objs.append(mobj)
+    return new_objs
 
 
 async def sync_preseem_model(preseem_model,
@@ -128,6 +164,7 @@ async def sync_preseem_model(preseem_model,
                              items,
                              packages=None,
                              accounts=None,
+                             services=None,
                              nm_model=None,
                              noact=False,
                              verbose=0):
@@ -150,10 +187,11 @@ async def sync_preseem_model(preseem_model,
         raise ValueError('source cannot be the empty string')
 
     async with sync_lock:
-        model_accounts, model_discs, model_sites, model_packages = await asyncio.gather(
+        model_accounts, model_discs, model_services, model_sites, model_packages = await asyncio.gather(
             preseem_model.Account.List(company_uuid=company_uuid),
             preseem_model.Element.ListElementDiscovery(company_uuid=company_uuid,
                                                        intent_role_uuid=role.bytes),
+            preseem_model.Service.List(company_uuid=company_uuid),
             preseem_model.Site.List(company_uuid=company_uuid),
             preseem_model.Package.List(company_uuid=company_uuid))
 
@@ -280,6 +318,7 @@ async def sync_preseem_model(preseem_model,
                     await call_and_log_error(
                         preseem_model.Element.UpdateElementDiscovery(
                             company_uuid=company_uuid, uuid=disc_uuid, inactive=True))
+        new_packages = []
         if packages is not None:
             assert isinstance(packages, dict)
             for pkg in packages.values():
@@ -293,16 +332,17 @@ async def sync_preseem_model(preseem_model,
                 if pkg.upload_rate and (not isinstance(pkg.upload_rate, int)
                                         or pkg.upload_rate < 0):
                     raise ValueError('invalid value for package up rate')
-            await _sync_objects(preseem_model,
-                                company_uuid,
-                                source,
-                                packages,
-                                model_packages,
-                                preseem_model.Package.Create,
-                                preseem_model.Package.Update,
-                                _package_field_map,
-                                noact=noact)
+            new_packages = await _sync_objects(preseem_model,
+                                               company_uuid,
+                                               source,
+                                               packages,
+                                               model_packages,
+                                               preseem_model.Package.Create,
+                                               preseem_model.Package.Update,
+                                               _package_field_map,
+                                               noact=noact)
 
+        new_accounts = []
         if accounts is not None:
             assert isinstance(accounts, dict)
             for account in accounts.values():
@@ -323,14 +363,78 @@ async def sync_preseem_model(preseem_model,
                 if account.last_name is not None and (not isinstance(
                         account.last_name, str) or account.last_name == ''):
                     raise ValueError('invalid string value for account last_name')
+            new_accounts = await _sync_objects(preseem_model,
+                                               company_uuid,
+                                               source,
+                                               accounts,
+                                               model_accounts,
+                                               preseem_model.Account.Create,
+                                               preseem_model.Account.Update,
+                                               _account_field_map,
+                                               noact=noact)
+
+        if services is not None:
+            assert isinstance(services, dict)
+            # Setup maps to lookup account and package UUIDs from the source IDs.
+            # We have to handle cases where there are multiple matching objects.
+            accounts_by_id = {}
+            for account in model_accounts + new_accounts:
+                al = accounts_by_id.get(account.source_id)
+                if al is None:
+                    accounts_by_id[account.source_id] = account
+                elif isinstance(al, list):
+                    al.append(account)
+                else:
+                    accounts_by_id[account.source_id] = [al, account]
+            packages_by_id = {}
+            for package in model_packages + new_packages:
+                pl = packages_by_id.get(package.source_id)
+                if pl is None:
+                    packages_by_id[package.source_id] = package
+                elif isinstance(pl, list):
+                    pl.append(package)
+                else:
+                    packages_by_id[package.source_id] = [pl, package]
+
+            # type/value checks are done in the source_model class
+            for service in services.values():
+                assert isinstance(service, Service)
+                account = accounts_by_id.get(
+                    service.account_id) if service.account_id else None
+                if account:
+                    if isinstance(account, list):  # multiple matching accounts
+                        acc = next((x for x in account if x.source == source), None)
+                        if acc:
+                            account = acc  # use account with matching source
+                        else:
+                            account = None  # multiple non-source matches; raise error
+                            service.err(
+                                'account_id', service.account_id,
+                                ValueError("Multiple accounts match the account ID"))
+                    if account:
+                        service._account_uuid = account.uuid
+                package = packages_by_id.get(
+                    service.package_id) if service.package_id else None
+                if package:
+                    if isinstance(package, list):  # multiple matching packages
+                        pkg = next((x for x in package if x.source == source), None)
+                        if pkg:
+                            package = pkg  # use package with matching source
+                        else:
+                            package = None  # multiple non-source matches; raise error
+                            service.err(
+                                'package_id', service.package_id,
+                                ValueError("Multiple packages match the package ID"))
+                    if package:
+                        service._package_uuid = package.uuid
             await _sync_objects(preseem_model,
                                 company_uuid,
                                 source,
-                                accounts,
-                                model_accounts,
-                                preseem_model.Account.Create,
-                                preseem_model.Account.Update,
-                                _account_field_map,
+                                services,
+                                model_services,
+                                preseem_model.Service.Create,
+                                preseem_model.Service.Update,
+                                _service_field_map,
                                 noact=noact)
 
         return uuid_map
